#ifdef CONFIG_RTNET_RTL
#define __NO_VERSION__
#endif
/*
 * 	NET3	Protocol independent device support routines.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *	Derived from the non IP parts of dev.c 1.0.19
 * 		Authors:	Ross Biro, <bir7@leland.Stanford.Edu>
 *				Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *				Mark Evans, <evansmp@uhura.aston.ac.uk>
 *
 *	Additional Authors:
 *		Florian la Roche <rzsfl@rz.uni-sb.de>
 *		Alan Cox <gw4pts@gw4pts.ampr.org>
 *		David Hinds <dhinds@allegro.stanford.edu>
 *		Alexey Kuznetsov <kuznet@ms2.inr.ac.ru>
 *		Adam Sulmicki <adam@cfar.umd.edu>
 *
 *	Changes:
 *		Alan Cox	:	device private ioctl copies fields back.
 *		Alan Cox	:	Transmit queue code does relevant stunts to
 *					keep the queue safe.
 *		Alan Cox	:	Fixed double lock.
 *		Alan Cox	:	Fixed promisc NULL pointer trap
 *		????????	:	Support the full private ioctl range
 *		Alan Cox	:	Moved ioctl permission check into drivers
 *		Tim Kordas	:	SIOCADDMULTI/SIOCDELMULTI
 *		Alan Cox	:	100 backlog just doesn't cut it when
 *					you start doing multicast video 8)
 *		Alan Cox	:	Rewrote net_bh and list manager.
 *		Alan Cox	: 	Fix ETH_P_ALL echoback lengths.
 *		Alan Cox	:	Took out transmit every packet pass
 *					Saved a few bytes in the ioctl handler
 *		Alan Cox	:	Network driver sets packet type before calling netif_rx. Saves
 *					a function call a packet.
 *		Alan Cox	:	Hashed net_bh()
 *		Richard Kooijman:	Timestamp fixes.
 *		Alan Cox	:	Wrong field in SIOCGIFDSTADDR
 *		Alan Cox	:	Device lock protection.
 *		Alan Cox	: 	Fixed nasty side effect of device close changes.
 *		Rudi Cilibrasi	:	Pass the right thing to set_mac_address()
 *		Dave Miller	:	32bit quantity for the device lock to make it work out
 *					on a Sparc.
 *		Bjorn Ekwall	:	Added KERNELD hack.
 *		Alan Cox	:	Cleaned up the backlog initialise.
 *		Craig Metz	:	SIOCGIFCONF fix if space for under
 *					1 device.
 *	    Thomas Bogendoerfer :	Return ENODEV for dev_open, if there
 *					is no device open function.
 *		Andi Kleen	:	Fix error reporting for SIOCGIFCONF
 *	    Michael Chastain	:	Fix signed/unsigned for SIOCGIFCONF
 *		Cyrus Durgin	:	Cleaned for KMOD
 *		Adam Sulmicki   :	Bug Fix : Network Device Unload
 *					A network device unload needs to purge
 *					the backlog queue.
 *	Paul Rusty Russel	:	SIOCSIFNAME
 */

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/notifier.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/rtnetlink.h>
//#include <net/slhc.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
//#include <net/br.h>
#include <net/dst.h>
#include <net/pkt_sched.h>
#include <net/profile.h>
#include <linux/init.h>
#include <linux/kmod.h>
#ifdef CONFIG_NET_RADIO
#include <linux/wireless.h>
#endif	/* CONFIG_NET_RADIO */
#ifdef CONFIG_PLIP
extern int plip_init(void);
#endif
#include <rtnet/rtnet.h>

#define RT

struct rt_device *rt_devices;
struct rtsocket *rt_protocols;

/*
 *	The list of packet types we will receive (as opposed to discard)
 *	and the routines to invoke.
 *
 *	Why 16. Because with 16 the only overlap we get on a hash of the
 *	low nibble of the protocol value is RARP/SNAP/X.25. 
 *
 *		0800	IP
 *		0001	802.3
 *		0002	AX.25
 *		0004	802.2
 *		8035	RARP
 *		0005	SNAP
 *		0805	X.25
 *		0806	ARP
 *		8137	IPX
 *		0009	Localtalk
 *		86DD	IPv6
 */


/*
 *	Device drivers call our routines to queue packets here. We empty the
 *	queue in the bottom half handler.
 */

static struct sk_buff_head backlog;


/******************************************************************************************

		Protocol management and registration routines

*******************************************************************************************/

/*
 *	Add a protocol ID to the list. Now that the input handler is
 *	smarter we can dispense with all the messy stuff that used to be
 *	here.
 *
 *	BEWARE!!! Protocol handlers, mangling input packets,
 *	MUST BE last in hash buckets and checking protocol handlers
 *	MUST start from promiscous ptype_all chain in net_bh.
 *	It is true now, do not change it.
 *	Explantion follows: if protocol handler, mangling packet, will
 *	be the first on list, it is not able to sense, that packet
 *	is cloned and should be copied-on-write, so that it will
 *	change it and subsequent readers will get broken packet.
 *							--ANK (980803)
 */
 
void rt_dev_add_pack(struct rt_device *rtdev,struct rtsocket *sock)
{
	sock->next=rt_protocols;
	rt_protocols=sock;
}


/*
 *	Remove a protocol ID from the list.
 */
 
void rt_dev_remove_pack(struct packet_type *pt)
{
	printk("rtnet: bug: dev_remove_pack() not implemented\n");
}


/*****************************************************************************************

			    Device Interface Subroutines

******************************************************************************************/

/* 
 *	Find an interface by name.
 */
 
struct rt_device *rt_dev_get(const char *name)
{
	return rt_dev_get_by_dev(dev_get(name));
}

struct rt_device * rt_dev_get_by_index(int ifindex)
{
	return rt_dev_get_by_dev(dev_get_by_index(ifindex));
}

struct rt_device * rt_dev_get_by_dev(struct device *dev)
{
	struct rt_device * rtdev;

	for(rtdev=rt_devices;rtdev;rtdev=rtdev->next){
		if(rtdev->dev==dev){
			return rtdev;
		}
	}
	return NULL;
}

struct rt_device *rt_dev_getbyhwaddr(unsigned short type, char *ha)
{
	struct rt_device * rtdev;
	struct device *dev;

	for(rtdev=rt_devices;rtdev;rtdev=rtdev->next){
		dev=rtdev->dev;
		if (dev->type == type &&
		    memcmp(dev->dev_addr, ha, dev->addr_len) == 0)
			return rtdev;
	}
	return NULL;
}


/*
 *	Create a new interface
 */
 
struct rt_device *rt_dev_alloc(struct device *dev)
{
	struct rt_device *rtdev;

	rtdev=kmalloc(sizeof(*rtdev),GFP_KERNEL);
	if(!rtdev)return NULL;
	memset(rtdev,0,sizeof(*rtdev));

	rtdev->dev=dev;

	rtdev->next=rt_devices;
	rt_devices=rtdev;

	rtdev->rx_sock=rtsocket_alloc();
	rtdev->tx_sock=rtsocket_alloc();

	return rtdev;
}


/*
 *	Prepare an interface for use. 
 */
 
int rt_dev_open(struct rt_device *rtdev)
{
	int ret = 0;
	struct device *dev=rtdev->dev;

	/*
	 *	Is it already up?
	 */

	if (dev->flags&IFF_UP)
		return 0;

	/*
	 *	Call device private open method
	 */
	 
	if (rtdev->open) 
  		ret = rtdev->open(rtdev);

	/*
	 *	If it went open OK then:
	 */
	 
	if (ret == 0) 
	{
		/*
		 *	nil rebuild_header routine,
		 *	that should be never called and used as just bug trap.
		 */

#if 0
		if (dev->rebuild_header == NULL)
			dev->rebuild_header = default_rebuild_header;
#endif

		/*
		 *	Set the flags.
		 */
		dev->flags |= (IFF_UP | IFF_RUNNING);

#if 0
		/*
		 *	Initialize multicasting status 
		 */
		dev_mc_upload(dev);
#endif

#if 0
		/*
		 *	Wakeup transmit queue engine
		 */
		dev_activate(dev);
#endif

#if 0
		/*
		 *	... and announce new interface.
		 */
		notifier_call_chain(&netdev_chain, NETDEV_UP, dev);
#endif

	}
	return(ret);
}

/*
 *	Completely shutdown an interface.
 */
 
int rt_dev_close(struct rt_device *rtdev)
{
	struct device *dev=rtdev->dev;

	if (!(rtdev->dev->flags&IFF_UP))
		return 0;

#if 0
	dev_lock_wait();
#endif

	/*
	 *	Call the device specific close. This cannot fail.
	 *	Only if device is UP
	 */
	 
	if (rtdev->stop)
		rtdev->stop(rtdev);

#if 0
	if (dev->start)
		printk("dev_close: bug %s still running\n", dev->name);
#endif

#if 0
	/*
	 *	Device is now down.
	 */
	dev_clear_backlog(dev);
#endif

	rtdev->dev->flags&=~(IFF_UP|IFF_RUNNING);

#if 0
	/*
	 *	Tell people we are going down
	 */
	notifier_call_chain(&netdev_chain, NETDEV_DOWN, dev);
#endif

	return(0);
}


#if 0
/*
 *	Device change register/unregister. These are not inline or static
 *	as we export them to the world.
 */

int register_netdevice_notifier(struct notifier_block *nb)
{
	return notifier_chain_register(&netdev_chain, nb);
}

int unregister_netdevice_notifier(struct notifier_block *nb)
{
	return notifier_chain_unregister(&netdev_chain,nb);
}
#endif

#if 0
/*
 *	Fast path for loopback frames.
 */
 
void dev_loopback_xmit(struct sk_buff *skb)
{
	struct sk_buff *newskb=skb_clone(skb, GFP_ATOMIC);
	if (newskb==NULL)
		return;

	newskb->mac.raw = newskb->data;
	skb_pull(newskb, newskb->nh.raw - newskb->data);
	newskb->pkt_type = PACKET_LOOPBACK;
	newskb->ip_summed = CHECKSUM_UNNECESSARY;
	if (newskb->dst==NULL)
		printk(KERN_DEBUG "BUG: packet without dst looped back 1\n");
	netif_rx(newskb);
}
#endif

int rt_dev_queue_xmit(struct sk_buff *skb)
{
	struct rt_device *rtdev = (struct rt_device *)skb->dev;
	int ret;

#if 0
rt_printk("rt_dev_queue_xmit()\n");
#endif
	ret=rtdev->xmit(skb,rtdev);
if(ret)rt_printk("xmit returned %d not 0\n",ret);

	return 0;
}


/*=======================================================================
			Receiver rotutines
  =======================================================================*/


/*
 *	Receive a packet from a device driver and queue it for the upper
 *	(protocol) levels.  It always succeeds. 
 */

void rtnetif_rx(struct sk_buff *skb)
{
	struct rt_device *rtdev;

	rtdev=rt_dev_get_by_dev(skb->dev);
	if(rtdev){
		skb->dev=(void *)rtdev;
	}

#if 0
	if(skb->stamp.tv_sec==0)
		get_fast_time(&skb->stamp);
#endif

	if(skb_queue_len(&backlog)>=20){
		rt_printk("dropping packet in rtnetif_rx()\n");
//		kfree_rtskb(skb);
	}
	if(skb){
		__skb_queue_tail(&backlog,skb);

#ifdef CONFIG_RTNET_TASK
		rtnet_mark_rxtask();
#else
		rtnet_bh();
#endif
	}else{
		rt_printk("called rtnetif_rx() with skb=NULL\n");
	}
}


void rtnetif_wake_queue(struct device *dev)
{
#ifdef CONFIG_RTNET_TASK
	//rtnet_mark_txtask();
#endif
}

/*
 *	When we are called the queue is ready to grab, the interrupts are
 *	on and hardware can interrupt and queue to the receive queue as we
 *	run with no problems.
 *	This is run as a bottom half after an interrupt handler that does
 *	mark_bh(NET_BH);
 */

void rtnet_bh(void)
{
	struct rtsocket *ptype;
	struct rtsocket *pt_prev;
	unsigned short type;

	/*
	 *	Can we send anything now? We want to clear the
	 *	decks for any more sends that get done as we
	 *	process the input. This also minimises the
	 *	latency on a transmit interrupt bh.
	 */

	/* XXX fix */
#if 0
	if (qdisc_head.forw != &qdisc_head)
		qdisc_run_queues();
#endif
  
	/*
	 *	Any data left to process. This may occur because a
	 *	mark_bh() is done after we empty the queue including
	 *	that from the device which does a mark_bh() just after
	 */

	/*
	 *	While the queue is not empty..
	 *
	 *	Note that the queue never shrinks due to
	 *	an interrupt, so we can do this test without
	 *	disabling interrupts.
	 */

//rt_printk("backlog %d\n",skb_queue_len(&backlog));
	while (!skb_queue_empty(&backlog)) 
	{
		struct sk_buff * skb;

		/*
		 *	We have a packet. Therefore the queue has shrunk
		 */
		skb = __skb_dequeue(&backlog);


		/*
	 	 *	Bump the pointer to the next structure.
		 * 
		 *	On entry to the protocol layer. skb->data and
		 *	skb->nh.raw point to the MAC and encapsulated data
		 */

		/* XXX until we figure out every place to modify.. */
		skb->h.raw = skb->nh.raw = skb->data;

		if (skb->mac.raw < skb->head || skb->mac.raw > skb->data) {
			rt_printk(KERN_CRIT "%s: wrong mac.raw ptr, proto=%04x\n", skb->dev->name, skb->protocol);
			kfree_rtskb(skb);
			continue;
		}

		/*
		 *	We got a packet ID.  Now loop over the "known protocols"
		 * 	list. There are two lists. The ptype_all list of taps (normally empty)
		 *	and the main protocol list which is hashed perfectly for normal protocols.
		 */

		type = skb->protocol;

		/* since we only have a few protocols, we won't bother to
		   use the hash */

		pt_prev = NULL;
		for (ptype = rt_protocols; ptype!=NULL; ptype=ptype->next)
		{
			if (ptype->proto==skb->protocol) {
/* XXX check device */
				if(pt_prev)
				{
					struct sk_buff *skb2=rtskb_clone(skb, GFP_ATOMIC);
					if(skb2)
						pt_prev->deliver(pt_prev,skb2);
				}
				pt_prev=ptype;
			}
		}

		/*
		 *	Is there a last item to send to ?
		 */

		if(pt_prev){
			pt_prev->deliver(pt_prev,skb);
		/*
		 * 	Has an unknown packet has been received ?
		 */
	 
		} else {
		/*	rt_printk("unknown packet\n"); */
			kfree_rtskb(skb);
		}
  	}	/* End of queue loop */
  	
  	/*
  	 *	We have emptied the queue
  	 */
	
	return;
}




#if 0

/* XXX this would be nice... -ds */

void dev_set_promiscuity(struct device *dev, int inc)
{
	unsigned short old_flags = dev->flags;

	dev->flags |= IFF_PROMISC;
	if ((dev->promiscuity += inc) == 0)
		dev->flags &= ~IFF_PROMISC;
	if (dev->flags^old_flags) {
#ifdef CONFIG_NET_FASTROUTE
		if (dev->flags&IFF_PROMISC) {
			netdev_fastroute_obstacles++;
			dev_clear_fastroute(dev);
		} else
			netdev_fastroute_obstacles--;
#endif
		dev_mc_upload(dev);
		printk(KERN_INFO "device %s %s promiscuous mode\n",
		       dev->name, (dev->flags&IFF_PROMISC) ? "entered" : "left");
	}
}
#endif



__initfunc(int rt_net_dev_init(void))
{
	skb_queue_head_init(&backlog);

	return 0;
}

