#ifdef CONFIG_RTNET_RTL
#define __NO_VERSION__
#endif
/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		The Internet Protocol (IP) output module.
 *
 * Version:	$Id: ip_output.c,v 1.2 2000/07/05 16:23:31 ds Exp $
 *
 * Authors:	Ross Biro, <bir7@leland.Stanford.Edu>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Donald Becker, <becker@super.org>
 *		Alan Cox, <Alan.Cox@linux.org>
 *		Richard Underwood
 *		Stefan Becker, <stefanb@yello.ping.de>
 *		Jorge Cwik, <jorge@laser.satlink.net>
 *		Arnt Gulbrandsen, <agulbra@nvg.unit.no>
 *
 *	See ip_input.c for original log
 *
 *	Fixes:
 *		Alan Cox	:	Missing nonblock feature in ip_build_xmit.
 *		Mike Kilburn	:	htons() missing in ip_build_xmit.
 *		Bradford Johnson:	Fix faulty handling of some frames when 
 *					no route is found.
 *		Alexander Demenshin:	Missing sk/skb free in ip_queue_xmit
 *					(in case if packet not accepted by
 *					output firewall rules)
 *		Mike McLagan	:	Routing by source
 *		Alexey Kuznetsov:	use new route cache
 *		Andi Kleen:		Fix broken PMTU recovery and remove
 *					some redundant tests.
 *	Vitaly E. Lavrov	:	Transparent proxy revived after year coma.
 *		Andi Kleen	: 	Replace ip_reply with ip_send_reply.
 *		Andi Kleen	:	Split fast and slow ip_build_xmit path 
 *					for decreased register pressure on x86 
 *					and more readibility. 
 *		Marc Boucher	:	When call_out_firewall returns FW_QUEUE,
 *					silently drop skb instead of failing with -EPERM.
 */

#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/init.h>

#include <net/snmp.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <net/arp.h>
#include <net/icmp.h>
#include <net/raw.h>
#include <net/checksum.h>
#include <linux/igmp.h>
#include <linux/ip_fw.h>
#include <linux/firewall.h>
#include <linux/mroute.h>
#include <linux/netlink.h>
#include <rtnet/rtnet.h>

#define RT


/* Generate a checksum for an outgoing IP datagram. */
static __inline__ void ip_send_check(struct iphdr *iph)
{
	iph->check = 0;
	iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);
}

/* 
 *		Add an ip header to a skbuff and send it out.
 */
void rt_ip_build_and_send_pkt(struct sk_buff *skb, struct sock *sk,
			   u32 saddr, u32 daddr, struct ip_options *opt)
{
	struct rt_rtable *rt = (struct rt_rtable *)skb->dst;
	struct iphdr *iph;
	struct rt_device *dev;
	
	/* Build the IP header. */
	if (opt)
		iph=(struct iphdr *)skb_push(skb,sizeof(struct iphdr) + opt->optlen);
	else
		iph=(struct iphdr *)skb_push(skb,sizeof(struct iphdr));

	iph->version  = 4;
	iph->ihl      = 5;
	iph->tos      = sk->ip_tos;
	iph->frag_off = 0;
#if 0
	if (ip_dont_fragment(sk, &rt->u.dst))
		iph->frag_off |= htons(IP_DF);
#else
	iph->frag_off |= htons(IP_DF);
#endif
	iph->ttl      = sk->ip_ttl;
	iph->daddr    = rt->rt_dst;
	iph->saddr    = rt->rt_src;
	iph->protocol = sk->protocol;
	iph->tot_len  = htons(skb->len);
	iph->id       = htons(ip_id_count++);
	skb->nh.iph   = iph;

	if (opt && opt->optlen) {
		iph->ihl += opt->optlen>>2;
		rt_ip_options_build(skb, opt, daddr, rt, 0);
	}

	dev = rt->rt_dev;

	ip_send_check(iph);

	/* Send it out. */
rt_printk("bogus rt_build_and_send_pkt()\n");
	//skb->dst->output(skb);
	return;
}



/*
 *	Fast path for unfragmented packets.
 */
int rt_ip_build_xmit(struct rtsocket *sk, 
		  int getfrag (const void *,
			       char *,
			       unsigned int,	
			       unsigned int),
		  const void *frag,
		  unsigned length,
		  struct ipcm_cookie *ipc,
		  struct rt_rtable *rt,
		  int flags)
{
	int err=0;
	struct sk_buff *skb;
	int df;
	struct iphdr *iph;
	struct rt_device *rtdev=rt->rt_dev;

	/*
	 *	Try the simple case first. This leaves fragmented frames, and by
	 *	choice RAW frames within 20 bytes of maximum size(rare) to the long path
	 */
#if 0
rt_printk("rt_ip_build_xmit()\n");
#endif

#if 0
	if (!sk->ip_hdrincl) {
#else
	if(1){
#endif
		length += sizeof(struct iphdr);

		/*
		 * 	Check for slow path.
		 */
#if 0
		if (length > rt->u.dst.pmtu || ipc->opt != NULL)  
			return ip_build_xmit_slow(sk,getfrag,frag,length,ipc,rt,flags); 
#endif
	} else {
		if (length > rtdev->dev->mtu) {
#if 0
			ip_local_error(sk, EMSGSIZE, rt->rt_dst, sk->dport, rtdev->mtu);
#endif
			return -EMSGSIZE;
		}
	}

	/*
	 *	Do path mtu discovery if needed.
	 */
	df = 0;
#if 0
	if (ip_dont_fragment(sk, &rt->u.dst))
		df = htons(IP_DF);
#else
		df = htons(IP_DF);
#endif

	/* 
	 *	Fast path for unfragmented frames without options. 
	 */ 
	{
	int hh_len = (rtdev->dev->hard_header_len + 15)&~15;

#if 0 /* XXX */
	skb = sock_alloc_send_skb(sk, length+hh_len+15,
				  0, flags&MSG_DONTWAIT, &err);
#endif
	skb = alloc_rtskb(length+hh_len+15,0);
	if(skb==NULL)
		goto error; 
	skb_reserve(skb, hh_len);
	}
	
#if 0
	skb->priority = sk->priority;
#else
	skb->priority = 0;
#endif
	skb->dst = (void *)rt; /* should clone */

	skb->nh.iph = iph = (struct iphdr *)skb_put(skb, length);
	
#if 0
	dev_lock_list();
#endif
	
#if 0
	if(!sk->ip_hdrincl) {
#else
	if(1) {
#endif
		iph->version=4;
		iph->ihl=5;
		iph->tos=sk->ip_tos;
		iph->tot_len = htons(length);
		iph->id=htons(ip_id_count++);
		iph->frag_off = df;
#if 0
		iph->ttl=sk->ip_mc_ttl;
		if (rt->rt_type != RTN_MULTICAST)
			iph->ttl=sk->ip_ttl;
#else
		iph->ttl=255;
#endif
		iph->protocol=sk->proto;
		iph->saddr=rt->rt_src;
		iph->daddr=rt->rt_dst;
		iph->check=0;
		iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);
		err = getfrag(frag, ((char *)iph)+iph->ihl*4,0, length-iph->ihl*4);
	}
	else
		err = getfrag(frag, (void *)iph, 0, length);

#if 0
	dev_unlock_list();
#endif

	if (err)
		goto error_fault;

	skb->dev=(struct device *)rtdev;
	if(!skb->dev){
		rt_printk("skb has bullshit device\n");
		return 0;
	}

	{
	char bull_mac[]={ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	if(!memcmp(bull_mac,rt->rt_dst_mac_addr,6)){
		rt_printk("skb has bullshit dst mac addr\n");
		return 0;
	}
	}

	rtdev->hard_header(skb,rtdev,ETH_P_IP,rt->rt_dst_mac_addr,rtdev->dev->dev_addr,
		skb->len);

	skb->dst=NULL;
	rtdev->xmit(skb,rtdev);
return 0;
	//return rt->u.dst.output(skb);

error_fault:
	err = -EFAULT;
	kfree_rtskb(skb);
error:
	ip_statistics.IpOutDiscards++;
	return err; 
}
		       


/*
 *	IP protocol layer initialiser
 */

static struct rtsocket *rt_ip_socket;



/*
 *	IP registers the packet type and then calls the subprotocol initialisers
 */

__initfunc(void rt_ip_init(void))
{
	rt_ip_socket=rtsocket_alloc();

	rt_ip_socket->deliver=rt_ip_rcv;
	rt_ip_socket->proto=htons(ETH_P_IP);

	rt_dev_add_pack(NULL,rt_ip_socket);

	rt_ip_rt_init();

#if 0
#ifdef CONFIG_PROC_FS
#ifdef CONFIG_IP_MULTICAST
	proc_net_register(&proc_net_igmp);
#endif
#endif	
#endif
}

