#ifdef CONFIG_RTNET_RTL
#define __NO_VERSION__
#endif
/*
    rtnet/module/rtnet/rtnet_dev.c
    RT network char device interface

    Linux RT networking
    Copyright (C) 1999 Zentropic Computing, LLC

    Author:	David Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/kmod.h>
#include <linux/miscdevice.h>

#include <asm/uaccess.h>

#include <rtnet/rtnet.h>


#define RTNET_MINOR 240

/* /dev/rtnet interface */

static int rtnet_open(struct inode *inode,struct file *file)
{
	return 0;
}

static int rtnet_ioctl(struct inode *inode,struct file *file,unsigned int cmd,unsigned long arg)
{
	struct rtnet_config cfg;
	struct rt_device *rtdev;
	int ret;

	if(!suser())return -EPERM;

	ret=copy_from_user(&cfg,(void *)arg,sizeof(cfg));
	if(ret)return -EFAULT;

	rtdev=rt_dev_get(cfg.if_name);
	if(!rtdev){
#ifdef CONFIG_KMOD
		ret=request_module(cfg.if_name);

		/* try again... */
		rtdev=rt_dev_get(cfg.if_name);
#endif

		if(!rtdev){
			printk("invalid interface %s\n",cfg.if_name);
			return -ENODEV;
		}
	}

	switch(cmd){
	case IOC_RT_IFUP:
		rt_dev_open(rtdev);

		rtdev->local_addr = cfg.ip_addr;

		rt_ip_route_add(rtdev,cfg.ip_netaddr,cfg.ip_mask);

		return 0;
	case IOC_RT_IFDOWN:
		rt_dev_close(rtdev);

		return 0;
	case IOC_RT_ROUTE_SOLICIT:
		rt_arp_solicit(rtdev,cfg.ip_addr);

		return 0;
	default:
		return -EINVAL;
	}
}

static struct file_operations rtnet_fops = {
	ioctl:	rtnet_ioctl,
	open:	rtnet_open,
};

static struct miscdevice rtnet_chr_misc_dev = {
	minor:		RTNET_MINOR,
	name:		"rtnet",
	fops:		&rtnet_fops,
};

void rtnet_chrdev_init(void)
{
	if(misc_register(&rtnet_chr_misc_dev) < 0) {
		printk("rtnet: unable to register rtnet char misc device\n");
	}
}


void rtnet_chrdev_cleanup(void)
{
	misc_deregister(&rtnet_chr_misc_dev);
}

